{% if lists_to_generate %}
T = TypeVar("T", covariant=True)

{% for list_type in lists_to_generate %}
class ListOf{{ list_type.name | capitalize }}(SubscriptionInstanceList[T]):
    min_items = {{ list_type.min_items }}
    {% if list_type.max_items -%}
    max_items = {{ list_type.max_items }}
    {% endif %}
{% endfor %}
{% endif -%}